import math
import humanfriendly
from pathlib import Path

__all__ = 'calc_gop', 'to_k', 'calc_buffer_size', 'calc_scale'


def calc_gop(bitrate, frame_rate, min_gop=-1):
    # https://github.com/vbence/stream-m#fragments
    return max(math.ceil((200 * 8000) / ((bitrate / 1000) * frame_rate)), min_gop)


def to_k(bytes):
    return f'{int(math.ceil(bytes / 1000))}k'


def calc_buffer_size(bitrate):
    if isinstance(bitrate, int):
        return to_k(bitrate * 1.5)
    return to_k(humanfriendly.parse_size(bitrate) * 1.5)


def calc_scale(size):
    in_w, in_h = size.split('x')
    return f'trunc({in_w}/2)*2:trunc({in_h}/2)*2'


'''
create_m3u8_master(presets.h264.H264xAnyDevice(), )
'''



# -*- coding: utf-8 -*-
from ...cli import *
from .._protocol import Protocol


__all__ = 'HlsHttpIn', 'HlsFileIn',


class HlsHttpIn(Protocol):
    """
    `HTTP Live Streaming protocol`
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    Read Apple HTTP Live Streaming compliant segmented stream as a uniform one.
    The M3U8 playlists describing the segments can be remote HTTP resources or
    local files, accessed using the standard file protocol. The nested protocol
    is declared by specifying "+proto" after the hls URI scheme name, where
    proto is either "file" or "http".

    |hls+http://host/path/to/remote/resource.m3u8|
    |hls+file://path/to/local/resource.m3u8|
    """
    def __init__(self, endpoint, *opt, **opts):
        super().__init__(
            *opt,
            Opt('i', 'endpoint', value='hls+http://' + endpoint),
            **opts
        )

class HlsFileIn(HlsHttpIn):
    def __init__(self, endpoint, *opt, **opts):
        super().__init__(
            *opt,
            Opt('i', 'endpoint', value='hls+http://' + endpoint),
            **opts
        )



# -*- coding: utf-8 -*-
from ..cli import *
from ._protocol import Protocol


__all__ = 'SrtpIn', 'SrtpOut',


class _Srtp(Protocol):
    """
    `Secure Real-time Transport Protocol`
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    The required syntax for an RTP URL is:
    |rtp://hostname[:port][?option=val...]|
    """


class SrtpIn(_Srtp):
    def __init__(self, endpoint, *opt, **opts):
        super().__init__(
            *opt,
            Opt('i', 'endpoint', value=endpoint),
            **opts
        )


class SrtpOut(_Srtp):
    def __init__(self, endpoint, *opt, **opts):
        super().__init__(
            *opt,
            Opt('', 'endpoint', value=endpoint),
            **opts
        )



# -*- coding: utf-8 -*-
from ..cli import *
from ._protocol import Protocol


__all__ = 'RtmpIn', 'RtmpOut',


DEFAULT_RTMP_OPT = [
    Opt('username'),
    Opt('password'),
    Opt('app'),
    Opt('playpath'),
    EnableOpt('listen'),
    IntOpt('timeout'),
    IntOpt('rtmp_buffer', 'buffer'),
    Opt('rtmp_conn', 'conn'),
    Opt('rtmp_flashver', 'flashver'),
    IntOpt('rtmp_flush_interval', 'flush_interval'),
    Opt('rtmp_live', 'live'),
    Opt('rtmp_playpath', 'playpath'),
    Opt('rtmp_subscribe', 'subscribe'),
    Opt('rtmp_swfhash', 'swfhash'),
    Opt('rtmp_swfverify', 'swfverify'),
    Opt('rtmp_swfsize', 'swfsize'),
    Opt('rtmp_swfurl', 'swfurl'),
    Opt('rtmp_tcurl', 'tcurl')
]


class _Rtmp(Protocol):
    """
    `Real-Time Messaging Protocol`
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    @username: (#str) An optional username (mostly for publishing).
    @password: (#str) An optional password (mostly for publishing).
    @app: (#str) It is the name of the application to access. It usually
        corresponds to the path where the application is installed on the RTMP
        server (e.g. /ondemand/, /flash/live/, etc.). You can override the
        value parsed from the URI through the rtmp_app option, too.
    @playpath: (#str) It is the path or name of the resource to play with
        reference to the application specified in app, may be prefixed by
        "mp4:". You can override the value parsed from the URI through the
        rtmp_playpath option, too.
    @listen: (#bool) Act as a server, listening for an incoming connection.
    @timeout: (#int) Maximum time to wait for the incoming connection. Implies
        listen.
    @buffer (rtmp_buffer): (#int) Set the client buffer time in milliseconds.
        The default is 3000.
    @conn (rtmp_conn): (#str) Extra arbitrary AMF connection parameters, parsed
        from a string, e.g. like B:1 S:authMe O:1 NN:code:1.23 NS:flag:ok O:0.
        Each value is prefixed by a single character denoting the type, B
        for Boolean, N for number, S for string, O for object, or Z for null,
        followed by a colon. For Booleans the data must be either 0 or 1 for
        FALSE or TRUE, respectively. Likewise for Objects the data must be 0 or
        1 to end or begin an object, respectively. Data items in subobjects may
        be named, by prefixing the type with ’N’ and specifying the name before
        the value (i.e. NB:myFlag:1). This option may be used multiple times to
        construct arbitrary AMF sequences.
    @flashver (rtmp_flashver): (#str) Version of the Flash plugin used to run
        the SWF player. The default is LNX 9,0,124,2. (When publishing, the
        default is FMLE/3.0 (compatible; <libavformat version>).)
    @flush_interval (rtmp_flush_interval): (#int) Number of packets flushed in
        the same request (RTMPT only). The default is 10.
    @live (rtmp_live): (#str) Specify that the media is a live stream. No
        resuming or seeking in live streams is possible. The default value is
        any, which means the subscriber first tries to play the live stream
        specified in the playpath. If a live stream of that name is not found,
        it plays the recorded stream. The other possible values are live and
        recorded.
    @pageurl (rtmp_pageurl): (#str) URL of the web page in which the media was
        embedded. By default no value will be sent.

    @playpath (rtmp_playpath): (#str) Stream identifier to play or to publish.
        This option overrides the parameter specified in the URI.
    @subscribe (rtmp_subscribe): (#str) Name of live stream to subscribe to.
        By default no value will be sent. It is only sent if the option is
        specified or if rtmp_live is set to live.
    @swfhash (rtmp_swfhash): (#str) SHA256 hash of the decompressed SWF
        file (32 bytes).
    @swfsize (rtmp_swfsize): (#str) Size of the decompressed SWF file,
        required for SWFVerification.
    @swfurl (rtmp_swfurl): (#str) URL of the SWF player for the media.
        By default no value will be sent.
    @swfverify (rtmp_swfverify): (#str) URL to player swf file, compute
        hash/size automatically.
    @tcurl (rtmp_tcurl): (#str) URL of the target stream. Defaults to
        proto://host[:port]/app.
    """
    OPT = DEFAULT_RTMP_OPT


class RtmpIn(_Rtmp):
    def __init__(self, endpoint, *opt, **opts):
        super().__init__(
            *opt,
            Opt('i', 'endpoint', value=endpoint),
            **opts
        )


class RtmpOut(_Rtmp):
    def __init__(self, endpoint, *opt, **opts):
        super().__init__(
            *opt,
            Opt('', 'endpoint', value=endpoint),
            **opts
        )



'''
AVC/H.264 Video Codec (Main or High profiles)
HE-AACv2 Audio Codec
Fragmented ISOBMFF Segments
MPEG-DASH “live” and “ondemand” profiles
Segments are IDR-aligned across Representations and have specific tolerances with regard to variation in segment durations.
CommonEncryption
SMPTE-TT Closed Captions


"urn:mpeg:dash:profile:isoff-live:2011". Supports ISOBMFF container, VOD and live streaming, using either the Segment Template: Time-Based or the Segment Template: Number-Based chunk addressing scheme.
"urn:mpeg:dash:profile:isoff-main:2011". Supports ISOBMFF container, VOD and live streaming, using the Segment List chunk addressing scheme.
'''
from ....codec import H264, Vpx, Vp9, MotionEstimation, BFrames, Aac, Opus
from ....output import Video, Audio
from ....format import Dash, Mp4
from .. import h264


#: DASH audio presets
#  MPEG-DASH spec specifically includes he-aac-v2 and multi-channel he-aac
LoFiDashAudio = Aac(bitrate='24k', sample_rate='44100', profile='aac_he_v2', channels=2)
MidFiDashAudio = Aac(bitrate='48k', sample_rate='44100', profile='aac_he_v2', channels=2)
HiFiDashAudio = Aac(bitrate='128k', sample_rate='44100', profile='aac_he_v2')


#: DASH recommended VPX settings
# http://wiki.webmproject.org/ffmpeg/vp9-encoding-guide
# http://wiki.webmproject.org/adaptive-streaming/instructions-to-playback-adaptive-webm-using-dash
# http://wiki.webmproject.org/zz-obsolete/instructions-to-playback-a-webm-dash-presentation

#: DASH
#  https://kvssoft.wordpress.com/2015/01/28/mpeg-dash-gop/
class BaseDashH264(h264.BaseH264):

    def _set_dash_opts(self):
        self.codec.key_interval(120)
        self.codec.flags('cgop')
        self.codec.min_key_interval(120)
        self.codec.scene_change_thresh(0)
        self.format = Mp4()
        self.format.movflags('frag_keyframe', 'faststart')
        self.format.merge(Dash())

        #if not self.format.min_seg_duration.val:
        #    self.format.min_seg_duration = 2

        self.format.single_file(True)
        self.audio = None
        self.disable_audio()
        self.disable_subtitles()


class H264xAnyDevice(h264.AnyDevice, BaseDashH264):

    def __init__(self, *opt, **opts):
        super(H264xAnyDevice, self).__init__(*opt, **opts)
        self._set_dash_opts()


class H264xAnyDevice360p(h264.AnyDevice360p, BaseDashH264):

    def __init__(self, *opt, **opts):
        super(H264xAnyDevice360p, self).__init__(*opt, **opts)
        self._set_dash_opts()


class H264x360p(h264.Main360p, BaseDashH264):

    def __init__(self, *opt, **opts):
        super(H264x360p, self).__init__(*opt, **opts)
        self._set_dash_opts()


class H264x480p(h264.High480p, BaseDashH264):

    def __init__(self, *opt, **opts):
        super(H264x480p, self).__init__(*opt, **opts)
        self._set_dash_opts()


class H264x576p(h264.High480p, BaseDashH264):

    def __init__(self, *opt, **opts):
        super(H264x576p, self).__init__(*opt, **opts)
        self._set_dash_opts()


class H264x720p(h264.High720p, BaseDashH264):

    def __init__(self, *opt, **opts):
        super(H264x720p, self).__init__(*opt, **opts)
        self._set_dash_opts()


'''
FRAGMENTS

Live streams consist of fragments (self-contained units of frames which are not referencing any frame outside the fragment). A fragment always starts with a key-frame (intra-frame) therefore it is important that the encoder put in key frames regularly as this determines the size of the fragments.

The ideal fragment size is around 200 kBytes (or 1600 kbits). The key frame interval can be calculated with this formula:

1600k / <bitrate> * <framerate>
e.g. if you are publishing a 500 kbit stream with 16 fps, then: 1600 / 500 * 16 = 51.2 (or 1600000 / 500000 * 16 = 51.2) so every 52nd video frame should be a key frame.

The server splits fragments when it seems necessary. A soft minimum for frame size is currently 100k (no new fragment is started if a new key frame arrives within 100 kBytes from a previous key frame).

The hard maximum for a fragments is 2048 kBytes. This is twice the size needed to hold 2 seconds of a 4096 kbit/sec HD stream.
'''



# -*- coding: utf-8 -*-
"""
`Theora`
============
libtheora wrapper for FFMPEG
"""
from ...cli import *
from .._codec import *


__all__ = ('Theora',)


class Theora(VideoCodec):
    OPT = [Opt('c:v', 'codec', value='libtheora')]



# -*- coding: utf-8 -*-
from ..cli import *
from ._container import *


__all__ = ('Aac',)


class Aac(Container):
    """
    ``AAC container format` `
    """
    OPT = [FlagsOpt('movflags', value='+faststart')]



# -*- coding: utf-8 -*-
"""
`H.265/HEVC`
============
libx265 wrapper for FFMPEG
"""
from ...cli import *
from .h264 import H264


__all__ = 'H265', 'Hevc',


class H265(H264):
    """
    ``H.265-specific Options``
    ~~~~~~~~~~~~~~~~~~~~~~~~~~
    @x265opts: (#dict(key=value)) Set any x265 option, see x265 --fullhelp
        for a list.
    """
    def __init__(self, *opt, **opts):
        super().__init__(
            *([KeyValOpt('x265opts', separator=':')] + list(opt)),
            **opts
        )


Hevc = H265
# nvenc_h265
# nvenc_hevc
# -pixel_format yuv444p
# https://trac.ffmpeg.org/wiki/HWAccelIntro



# -*- coding: utf-8 -*-
"""
`H.264`
=======

``iOS Compatability``
|‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾|
| Profile     Level      Devices                                               |
|––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––|
| Baseline    3.0   All devices                                                |
|                    -profile:v baseline -level 3.0                            |
| Baseline    3.1   iPhone 3G and later, iPod touch 2nd generation and later   |
|                    -profile:v baseline -level 3.1                            |
| Main        3.1   iPad (all versions), Apple TV 2 and later, iPhone 4 and    |
|                   later                                                      |
|                    -profile:v main -level 3.1                                |
| Main        4.0   Apple TV 3 and later, iPad 2 and later, iPhone 4s and later|
|                    -profile:v main -level 4.0                                |
| High        4.0   Apple TV 3 and later, iPad 2 and later, iPhone 4s and later|
|                    -profile:v high -level 4.0                                |
| High        4.1   iPad 2 and later, iPhone 4s and later, iPhone 5c and later |
|                    -profile:v high -level 4.1                                |
| High        4.2   iPad Air and later, iPhone 5s and later                    |
|                    -profile:v high -level 4.2                                |
‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾
Source: https://trac.ffmpeg.org/wiki/Encode/H.264


Presets
~~~~~~~
Preset determines how fast the encoding process will be – at the expense of
compression efficiency. Put differently, if you choose ultrafast, the encoding
process is going to run fast, but the file size will be larger when compared to
medium. The visual quality will be the same. Valid presets are ultrafast,
superfast, veryfast, faster, fast, medium, slow, slower, veryslow and placebo.


Encoding for dumb players
~~~~~~~~~~~~~~~~~~~~~~~~~
You may need to use |-pix_fmt yuv420p| for your output to work in QuickTime and
most other players. These players only supports the YUV planar color space with
4:2:0 chroma subsampling for H.264 video. Otherwise, depending on your source,
ffmpeg may output to a pixel format that may be incompatible with these
players.


Encoding for web
~~~~~~~~~~~~~~~~
|-movflags +faststart| (Container-level)

"""
from ...cli import *
from .._codec import *


__all__ = 'H264',


class H264(VideoCodec):
    """
    libx264 supports an impressive number of features, including 8x8 and 4x4
    adaptive spatial transform, adaptive B-frame placement, CAVLC/CABAC entropy
    coding, interlacing (MBAFF), lossless mode, psy optimizations for detail
    retention (adaptive quantization, psy-RD, psy-trellis).

    Many libx264 encoder options are mapped to FFmpeg global codec options,
    while unique encoder options are provided through private options.
    Additionally the x264opts and x264-params private options allows one to
    pass a list of key=value tuples as accepted by the libx264 x264_param_parse
    function.

    To reduce the duplication of documentation, only the private options and
    some others requiring special attention are documented here. For the
    documentation of the undocumented generic options, see :class:VideoCodec.

    |‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾|
    | Levels          Max Bitrates            Max Resolutions                  |
    |~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~|
    | Level     Baseline/Main    High      Resolution@Max Frame Rate           |
    |––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––|
    | 1             64            80           176×144@15.0                    |
    | 1.1           192           240          352×288@7.5                     |
    | 2             2,000         2,500        352×288@30.0                    |
    | 2.1           4,000         5,000        352×576@25.0                    |
    | 3             10,000        12,500       720×576@25.0                    |
    | 3.1           14,000        17,500       1,280×720@30.0                  |
    | 3.2           20,000        25,000       1,280×1,024@42.2                |
    | 4             20,000        25,000       2,048×1,024@30.0                |
    | 4.1           50,000        62,500       2,048×1,024@30.0                |
    | 4.2           50,000        62,500       2,048×1,080@60.0                |
    | 5             135,000       168,750      3,672×1,536@26.7                |
    | 5.1           240,000       300,000      4,096×2,304@26.7                |
    | 5.2           240,000       300,000      4,096×2,304@56.3                |
    ‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾

    ``H.264-specific options``
    ~~~~~~~~~~~~~~~~~~~~~~~~~~
    @profile: (#str) Set profile restrictions.
    @fast_firstpass (fastfirstpass): (#bool) Enable fast settings when encoding
        first pass, when set to 1. When set to 0, it has the same effect of
        x264’s --slow-firstpass option.
    @preset: (#str) Preset determines how fast the encoding process will be –
        at the expense of compression efficiency. Valid presets are ultrafast,
        superfast, veryfast, faster, fast, medium, slow, slower, veryslow and
        placebo.
    @crf: (#int) Set the quality for constant quality mode.
    @crf_max: (#int) In CRF mode, prevents VBV from lowering quality beyond
        this point.
    @qp: (#int) Set constant quantization rate control method parameter.
    @aq_mode (aq-mode): (#str) Set AQ method.
    @aq_strength (aq-strength): (#float) Set AQ strength, reduce blocking and
        blurring in flat and textured areas.
    @psy: (#int) Use psychovisual optimizations when set to 1. When set to 0,
        it has the same effect as x264’s --no-psy option.
    @psy_rd (psy-rd): (#tuple(#float)) Set strength of psychovisual
        optimization, in psy-rd:psy-trellis format.
    @rc_lookahead (rc-lookahead): (#int) Set number of frames to look ahead for
        frametype and ratecontrol.
    @weightb: (#int) Enable weighted prediction for B-frames when set to 1.
        When set to 0, it has the same effect as x264’s --no-weightb option.
    @weightp: (#str) Set weighted prediction method for P-frames.
    @ssim: (#bool) Enable calculation and printing SSIM stats after the
        encoding.
    @intra_refresh (intra-refresh): (#int) Enable the use of Periodic Intra
        Refresh instead of IDR frames when set to 1.
    @avcintra_class (avcintra-class): (#int) Configure the encoder to generate
        AVC-Intra. Valid values are 50, 100 and 200.
    @bluray_compat (bluray-compat): (#bool) Configure the encoder to be
        compatible with the bluray standard. It is a shorthand for setting
        "bluray-compat=1 force-cfr=1".
    @b_bias (b-bias): (#int) Set the influence on how often B-frames are used.
    @b_pyramid (b-pyramid): (#str) Set method for keeping of some B-frames as
        references.
    @mixed_refs (mixed-refs): (#int) Enable the use of one reference per
        partition, as opposed to one reference per macroblock when set to
        1. When set to 0, it has the same effect as x264’s --no-mixed-refs
        option.
    @8x8dct: (#int) Enable adaptive spatial transform (high profile 8x8
        transform) when set to 1. When set to 0, it has the same effect as
        x264’s --no-8x8dct option.
    @fast_pskip (fast-pskip): (#int) Enable early SKIP detection on
        P-frames when set to 1. When set to 0, it has the same effect as
        x264’s --no-fast-pskip option.
    @aud: (#int) Enable use of access unit delimiters when set to 1.
    @mbtree: (#int) Enable use macroblock tree ratecontrol when set to 1.
        When set to 0, it has the same effect as x264’s --no-mbtree option.
    @cplxblur: (#float) Set fluctuations reduction in QP (before curve
        compression).
    @partitions: (#tuple(#str)) e.g. ('parti4x4', 'partp8x8') or ('all',)
        Set partitions to consider as a comma-separated list of.
    @direct_prediction (direct-pred): (#str) Set direct MV prediction mode.
    @slice_max_size (slice-max-size): (#str) Set the limit of the size of each
        slice in bytes. If not specified but RTP payload size (ps) is specified,
        that is used.
    @nal_hrd (nal-hrd): (#str) Set signal HRD information (requires vbv-bufsize
        to be set).
    @x264opts: (#dict(key=value)) Set any x264 option, see x264 --fullhelp for
        a list.
    @deblock_beta: (#int)
    @deblock_alpha: (#int) One of H.264's main features is the in-loop
        deblocker, which avoids the problem of blocking artifacts disrupting
        motion estimation. This requires a small amount of decoding CPU, but
        considerably increases quality in nearly all cases. Its strength may
        be raised or lowered in order to avoid more artifacts or keep more
        detail, respectively. Deblock has two parameters: alpha (strength) and
        beta (threshold). Recommended defaults:-deblockalpha 0 -deblockbeta 0
        (Must have '-flags +loop')
    """
    OPT = [
        Opt('c:v', 'codec', value='libx264'),
        OneOfOpt('profile:v', 'profile',
                  choices=('baseline', 'extended', 'main', 'high')),
        OneOfOpt(
            'level',
            cast=str,
            choices=(
                '1', '1b', '1.1', '1.2', '1.3', '2', '2.1', '3', '3.1', '3.2',
                '4', '4.1', '4.2', '5', '5.1', '5.2', '6', '6.1', '6.2'
            )
        ),
        OneOfOpt('preset', cast=str,
                 choices=('ultrafast', 'superfast', 'veryfast',
                          'faster', 'fast', 'medium', 'slow', 'slower',
                          'veryslow', 'placebo')),
        BoundedOpt('crf', lower=0, upper=51),
        BoundedOpt('crf_max', lower=0, upper=51),
        BoundedOpt('qp', lower=0, upper=51),
        OneOfOpt('aq-mode', 'aq_mode',
                 choices=('none', 'variance', 'autovariance')),
        FloatOpt('aq-strength', 'aq_strength'),
        EnableOpt('psy'),
        GroupOpt('psy-rd', 'psy_rd', cast=float, separator=':'),
        IntOpt('rc-lookahead', 'rc_lookahead'),
        EnableOpt('coder'),
        EnableOpt('fastfirstpass', 'fast_firstpass'),
        EnableOpt('weightb'),
        OneOfOpt('weightp', choices=('none', 'simple', 'smart')),
        Flag('ssim'),
        EnableOpt('mixed-refs', 'mixed_refs'),
        EnableOpt('8x8dct'),
        EnableOpt('fast-pskip', 'fast_pskip'),
        EnableOpt('aud'),
        EnableOpt('mbtree'),
        FloatOpt('cplxblur'),
        EnableOpt('intra-refresh', 'intra_refresh'),
        OneOfOpt('avcintra-class', 'avcintra_class', cast=int,
                 choices=(50, 100, 200)),
        GroupOpt('partitions', separator=','),
        Flag('bluray-compat', 'bluray_compat'),
        ByteOpt('slice-max-size', 'slice_max_size'),
        OneOfOpt('b-pyramid', 'b_pyramid',
                 choices=('none', 'strict', 'normal')),
        OneOfOpt('thread_type', choices=('slice', 'frame')),
        OneOfOpt('nal-hrd', 'nal_hrd', choices=('none', 'vbr', 'cbr')),

        OneOfOpt('direct-pred', 'direct_prediction',
                 choices=('none', 'spatial', 'temporal', 'auto')),
        KeyValOpt('x264opts', separator=':'),
        Opt('tune'),
        IntOpt('deblockalpha', 'deblock_alpha'),
        IntOpt('deblockbeta', 'deblock_beta')
    ]


# nvenc_h264
# -pixel_format yuv444p
# https://trac.ffmpeg.org/wiki/HWAccelIntro



# -*- coding: utf-8 -*-
from ...cli import *
from .._container import *


__all__ = ('Segmenter',)


class Segmenter(Container):
    """
    This muxer outputs streams to a number of separate files of nearly fixed
    duration. Output filename pattern can be set in a fashion similar to image2,
    or by using a strftime template if the strftime option is enabled.

    |stream_segment| is a variant of the muxer used to write to streaming output
    formats, i.e. which do not require global headers, and is recommended for
    outputting e.g. to MPEG transport stream segments. |ssegment| is a shorter
    alias for |stream_segment|.

    Every segment starts with a keyframe of the selected reference stream,
    which is set through the |reference_stream| option.

    Note that if you want accurate splitting for a video file, you need to make
    the input key frames correspond to the exact splitting times expected by
    the segmenter, or the segment muxer will start the new segment with the key
    frame found next after the specified start time.

    The segment muxer works best with a single constant frame rate video.

    Optionally it can generate a list of the created segments, by setting the
    option segment_list. The list type is specified by the |segment_list_type|
    option. The entry filenames in the segment list are set by default to the
    basename of the corresponding segment files.


    ``Basic stream segmenter options``
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    @incr_timecode (increment_tc): (#bool) If set to 1, increment timecode
        between each segment If this is selected, the input need to have a
        timecode in the first video stream. Default value is 0.
    @reference_stream: (#str) Set the reference stream, as specified by the
        string specifier. If specifier is set to auto, the reference is chosen
        automatically. Otherwise it must be a stream specifier (see the “Stream
        specifiers” chapter in the ffmpeg manual) which specifies the reference
        stream. The default value is auto.
    @inner_format (segment_format): (#str) Override the inner container format,
        by default it is guessed by the filename extension.
    @inner_format_opt (segment_format_options): (#dict) Set output format
        options using a :-separated list of key=value parameters. Values
        containing the : special character must be escaped.
    @list_file (segment_list): (#str) Generate also a listfile named name. If
        not specified no listfile is generated.
    @list_flags (segment_list_flags): (#tuple(flags)) Set flags affecting the
        segment list generation.
    @list_size (segment_list_size): (#int) Update the list file so that it
        contains at most size segments. If 0 the list file will contain all the
        segments. Default value is 0.
    @list_entry_prefix (segment_list_entry_prefix): (#str) Prepend prefix to
        each entry. Useful to generate absolute paths. By default no prefix is
        applied.
    @list_type (segment_list_type): (#str) Listing format: 'flat', 'csv, ext',
        'ffconcat', 'm3u8'
    @time (segment_time): (#int) Set segment duration to time, the value must
        be a duration specification. Default value is "2". See also the
        segment_times option.
    @at_clock_time (segment_atclocktime): (#bool) If set to "1" split at regular
        clock time intervals starting from 00:00 o’clock. The time value
        specified in segment_time is used for setting the length of the
        splitting interval.
    @clocktime_offset (segment_clocktime_offset): (#int) Delay the segment
        splitting times with the specified duration when using
        segment_atclocktime.
    @clocktime_wrap_duration (segment_clocktime_wrap_duration): (#float) Force
        the segmenter to only start a new segment if a packet reaches the muxer
        within the specified duration after the segmenting clock time. This way
        you can make the segmenter more resilient to backward local time jumps,
        such as leap seconds or transition to standard time from daylight
        savings time.
    @time_delta (segment_time_delta): (#float) Specify the accuracy time when
        selecting the start time for a segment, expressed as a duration
        specification. Default value is "0".
    @times (segment_times): (#tuple(#float)) Specify a list of split points.
        Times contains a list of comma separated duration specifications, in
        increasing order.
    @frames (segment_frames): (#tuple(#int)) Specify a list of split video
        frame numbers. frames contains a list of comma separated integer
        numbers, in increasing order.
    @wrap (segment_wrap): (#int) Wrap around segment index once it reaches
        limit.
    @start_number (segment_start_number): (#int) Set the sequence number of the
        first segment. Defaults to 0.
    @strftime: (#bool) Use the strftime function to define the name of the new
        segments to write. If this is selected, the output segment name must
        contain a strftime function template. Default value is 0.
    @break_non_keyframes: (#bool) If enabled, allow segments to start on
        frames other than keyframes. This improves behavior on some players
        when the time between keyframes is inconsistent, but may make things
        worse on others, and can cause some oddities during seeking.
        Defaults to 0.
    @reset_timestamps: (#bool) Reset timestamps at the begin of each segment,
        so that each segment will start with near-zero timestamps. It is meant
        to ease the playback of the generated segments. May not work with some
        combinations of muxers/codecs. It is set to 0 by default.
    @initial_offset: (#float) Specify timestamp offset to apply to the output
        packet timestamps. The argument must be a time duration specification,
        and defaults to 0.
    @write_empty (write_empty_segments): (#bool) If enabled, write an empty
        segment if there are no packets during the period a segment would
        usually span. Otherwise, the segment will be filled with the next
        packet written. Defaults to 0.
    """
    OPT = [
        Opt('f', 'format', value='segment'),
        EnableOpt('increment_tc', 'incr_timecode'),
        Opt('reference_stream'),
        Opt('segment_format', 'inner_format'),
        KeyValOpt('segment_format_options', 'inner_format_opt',
                  separator=':'),
        Opt('segment_list', 'list_file'),
        FlagsOpt('segment_list_flags', 'list_flags'),
        IntOpt('segment_list_size', 'list_size'),
        Opt('segment_list_entry_prefix', 'list_entry_prefix'),
        OneOfOpt('segment_list_type', 'list_type',
                 choices=('flat', 'csv, ext', 'ffconcat', 'm3u8')),
        FloatOpt('segment_time', 'time'),
        EnableOpt('segment_atclocktime', 'at_clock_time'),
        FloatOpt('segment_clocktime_offset', 'clocktime_offset'),
        FloatOpt('segment_clocktime_wrap_duration',
               'clocktime_wrap_duration'),
        FloatOpt('segment_time_delta', 'time_delta'),
        GroupOpt('segment_times', 'times', cast=float, separator=','),
        GroupOpt('segment_frames', 'frames', cast=int, separator=','),
        IntOpt('segment_wrap', 'wrap'),
        IntOpt('segment_start_number', 'start_number'),
        EnableOpt('strftime'),
        EnableOpt('break_non_keyframes'),
        EnableOpt('reset_timestamps'),
        FloatOpt('initial_offset'),
        EnableOpt('write_empty_segments', 'write_empty')
    ]



# -*- coding: utf-8 -*-
from ...cli import *
from .._container import *


__all__ = ('SmoothStreaming',)


class SmoothStreaming(Container):
    """
    Smooth Streaming muxer generates a set of files (Manifest, chunks) suitable
    for serving with conventional web server.


    ``SmoothStreaming muxer options``
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    @window_size: (#int) Specify the number of fragments kept in the manifest.
        Default 0 (keep all).
    @extra_window_size: (#int) Specify the number of fragments kept outside of
        the manifest before removing from disk. Default 5.
    @lookahead_count: (#int) Specify the number of lookahead fragments.
        Default 2.
    @min_frag_duration: (#int) Specify the minimum fragment duration
        (in microseconds). Default 5000000.
    @remove_at_exit: (#int) Specify whether to remove all fragments when
        finished. Default 0 (do not remove).
    """
    OPT = [
        Opt('f', 'format', value='smoothstreaming'),
        IntOpt('window_size'),
        IntOpt('extra_window_size'),
        IntOpt('lookahead_count'),
        IntOpt('min_frag_duration'),
        EnableOpt('remove_at_exit')
    ]



# -*- coding: utf-8 -*-
from ..cli import *
from ._container import *


__all__ = 'Matroska', 'WebM',


class Matroska(Container):
    """
    `Matroska (mkv) container format`
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    @reserve_index_space: (#int) By default, this muxer writes the index for
        seeking (called cues in Matroska terms) at the end of the file,
        because it cannot know in advance how much space to leave for the
        index at the beginning of the file. However for some use cases – e.g.
        streaming where seeking is possible but slow – it is useful to put the
        index at the beginning of the file.
    """
    OPT = [
        Opt('f', 'format', value='mkv'),
        IntOpt('reserve_index_space')
    ]


class WebM(Matroska):
    OPT = [Opt('f', 'format', value='webm')]



from ..output import Video
from ..codec import VideoCodec
from ..protocol import FileOut


class Thumbnails(Video):
    def __init__(self, *opt, filename='%d.jpg', fps=1, **opts):
        super().__init__(
            *opt,
            codec=VideoCodec(),
            output=FileOut(filename),
            filters={'fps': fps},
            **opts
        )



#: FastStart exceptions
class FastStartError(Exception):
    """ Standard exception """
    pass


class AlreadyMoved(FastStartError):
    """ Raised when the input's moov atom is already in the correct place. """
    pass


class FileIsCorrupt(FastStartError):
    """ Raised when the input has a perceived corruption. For example,
        unreadable atoms
    """
    pass


class FileIsUnsupported(FastStartError):
    """ Raised when the moov atom is compressed. """
    pass


class DasherError(Exception):
    pass



# -*- coding: utf-8 -*-
from ..cli import *
from ._protocol import Protocol


__all__ = 'HttpIn', 'HttpOut',


HTTP_OPT = [
    BoundedOpt('seekable', lower=-1, upper=1),
    EnableOpt('chunked_post', 'chunked'),
    Opt('content_type'),
    Opt('proxy'),
    Opt('headers'),
    EnableOpt('multiple_requests'),
    Opt('post_data'),
    Opt('user_agent'),
    IntOpt('timeout'),
    Flag('reconnect_at_eof'),
    Flag('reconnect_streamed'),
    IntOpt('reconnect_delay_max'),
    Opt('mime_type'),
    EnableOpt('icy'),
    Opt('icy_metadata_headers'),
    Opt('icy_metadata_packet'),
    Opt('cookies'),
    IntOpt('offset'),
    IntOpt('end_offset'),
    Opt('method'),
    IntOpt('listen')
]


class _Http(Protocol):
    """
    `HTTP access protocol`
    ~~~~~~~~~~~~~~~~~~~~~~
    @seekable: (#int) Control seekability of connection. If set to 1 the
        resource is supposed to be seekable, if set to 0 it is assumed not to
        be seekable, if set to -1 it will try to autodetect if it is seekable.
        Default value is -1.
    @chunked (chunked_post): (#bool) If set to 1 use chunked Transfer-Encoding
        for posts, default is 1.
    @content_type: (#str) Set a specific content type for the POST messages.
    @proxy (http_proxy): (#str) @et HTTP proxy to tunnel through e.g.
        http://example.com:1234
    @headers: (#str) Set custom HTTP headers, can override built in default
        headers. The value must be a string encoding the headers.
    @multiple_requests: (#bool) Use persistent connections if set to 1, default
        is 0.
    @post_data: (#str) Set custom HTTP post data.
    @user_agent: (#str) Override the User-Agent header. If not specified the
        protocol will use a string describing the libavformat build.
    @timeout: (#int) Set timeout in microseconds of socket I/O operations used
        by the underlying low level operation. By default it is set to -1,
        which means that the timeout is not specified.
    @reconnect_at_eof: (#bool) If set then eof is treated like an error and
        causes reconnection, this is useful for live / endless streams.
    @reconnect_streamed: (#bool) If set then even streamed/non seekable streams
        will be reconnected on errors.
    @reconnect_delay_max: (#int) Sets the maximum delay in seconds after which
        to give up reconnecting
    @mime_type: (#str) Export the MIME type.
    @icy: (#bool) If set to 1 request ICY (SHOUTcast) metadata from the server.
        If the server supports this, the metadata has to be retrieved by the
        application by reading the icy_metadata_headers and icy_metadata_packet
        options. The default is 1.
    @icy_metadata_headers: (#str) If the server supports ICY metadata, this
        contains the ICY-specific HTTP reply headers, separated by newline
        characters.
    @icy_metadata_packet: (#str) If the server supports ICY metadata, and
        icy was set to 1, this contains the last non-empty metadata packet
        sent by the server. It should be polled in regular intervals by
        applications interested in mid-stream metadata updates.
    @cookies: (#str) Set the cookies to be sent in future requests. The format
        of each cookie is the same as the value of a Set-Cookie HTTP response
        field. Multiple cookies can be delimited by a newline character.
    @offset: (#int) Set initial byte offset
    @end_offset: (#int) Try to limit the request to bytes preceding this offset.
    @method: (#str) When used as a client option it sets the HTTP method for
        the request. When used as a server option it sets the HTTP method that
        is going to be expected from the client(s). If the expected and the
        received HTTP method do not match the client will be given a Bad
        Request response. When unset the HTTP method is not checked for now.
        This will be replaced by autodetection in the future.
    @listen: (#int) If set to 1 enables experimental HTTP server. This can be
        used to send data when used as an output option, or read data from a
        client with HTTP POST when used as an input option. If set to 2 enables
        experimental mutli-client HTTP server. This is not yet implemented in
        ffmpeg.c or ffserver.c and thus must not be used as a command line
        option.
    """
    OPT = HTTP_OPT


class HttpIn(_Http):
    def __init__(self, endpoint, *opt, **opts):
        super().__init__(
            *opt,
            Opt('i', 'endpoint', value=endpoint),
            **opts
        )


class HttpOut(_Http):
    def __init__(self, endpoint, *opt, **opts):
        super().__init__(
            *opt,
            Opt('', 'endpoint', value=endpoint),
            **opts
        )



import os
import struct
import humanfriendly
from collections import namedtuple

import io

from .exceptions import AlreadyMoved, FileIsCorrupt, \
                        FileIsUnsupported, FastStartError


# Atom storage namedtuple
Atom = namedtuple('Atom', 'name position size')
REQUIRED_ATOMS = {'moov', 'mdat'}
ATOMS = {'ftyp', 'moov', 'moof', 'mdat', 'free', 'fourcc', 'mfra'}
CHILD_ANCESTORS = {"trak", "mdia", "minf", "stbl"}
DESIRED_CHILDREN = {"stco", "co64"}


class FastStart(object):
    __slots__ = ('cache', 'stream', 'indexes', '_stream_pos', 'BUFFER_SIZE')

    def __init__(self, buffer_size=24 * 1024 * 1024):
        """ @buffer_size: (#int) size of each stream chunk in bytes, defaults
                to 50M. The buffer size also must be a multiple of 16.
        """
        self.reset()
        self.BUFFER_SIZE = buffer_size
        if buffer_size < 16 or \
           float(buffer_size / 16) != float(buffer_size // 16):
            raise FastStartError('Buffer size must be at least 16 bytes and '
                                 'must be a multiple of 16. The recommended '
                                 'size is at least 10M.')

    def __repr__(self):
        return '<FastStart: -buffer_size=%s>' % \
               humanfriendly.format_size(self.BUFFER_SIZE)

    def get_indexes(self, chunk, chunk_size):
        """ Yields an index of top level atoms, their absolute byte-position in
            the file and their size
        """
        while(chunk):
            try:
                # Reads the next atom in the chunk
                atom = self.read_atom(chunk, self._stream_pos)
            except struct.error as e:
                break

            if atom.size == 0:
                if atom.name == "mdat":
                    # Some files may end in mdat with no size set, which
                    # generally means to seek to the end of the file
                    break
                else:
                    # Not necessarily malformed nor signaling anything,
                    # so just continue
                    continue
            # The next top-level atom will be found immediately after the end
            # of this atom, so that's where we will skip to
            skip_to = atom.position + atom.size
            chunk_end = self._stream_pos + chunk_size
            # Yield the newly discovered atom
            yield atom
            # If the next atom position is available within this chunk,
            # find it, otherwise break the loop and move on to the next chunk
            if skip_to < chunk_end:
                # We need at least 16 available bytes to unpack properly
                # so we fetch any extras we need from the stream
                if chunk_end - skip_to < 16:
                    # Read the required bytes from the stream
                    stream = self.stream.read(chunk_end - skip_to)
                    # Write the new bytes to the chunk AND the cached data
                    chunk.seek(0, os.SEEK_END)
                    chunk.write(stream)
                    self.cache.write(stream)
                # Seek the chunk to the next atom location
                chunk.seek(skip_to - self._stream_pos)
            else:
                break

    def read_atom(self, chunk, padding=0):
        """ Read an atom and return a tuple of (size, type) where size is the
            size in bytes (including the 8 bytes already read) and type is a
            "fourcc" like "ftyp" or "moov".
        """
        pos = chunk.tell()
        read = chunk.read(8)
        atom_size, atom_type = struct.unpack(">L4s", read)

        if atom_size == 1:
            atom = Atom(atom_type.decode('ascii', 'ignore'),
                        padding + pos,
                        struct.unpack(">Q", chunk.read(8))[0])
        else:
            atom = Atom(atom_type.decode('ascii', 'ignore'),
                        padding + pos,
                        atom_size)
        return atom

    def _check_compressed(self, moov_atom):
        is_compressed = self._moov_is_compressed(moov_atom)
        if is_compressed:
            raise FileIsUnsupported("Files with compressed headers are not "
                                    "supported")

    def _moov_is_compressed(self, moov_atom):
        """ Scan the atoms under the moov atom and detect whether or not the
            atom data is compressed.
        """
        # Seek to the beginning of the moov atom contents (8 skips the atom
        # type identifier)
        self.cache.seek(moov_atom.position + 8)

        # Step through the moov atom children to see if a cmov atom is
        # among them
        stop = moov_atom.position + moov_atom.size

        while self.cache.tell() < stop:
            child_atom = self.read_atom(self.cache)
            self.cache.seek(self.cache.tell() + child_atom.size - 8)
            # cmov means there is a compressed moov header
            if child_atom.name == 'cmov':
                return True

        return False

    def _patch_moov(self, atom, offset):
        # Seek to the beginning of the moov atom within the cache
        self.cache.seek(atom.position)
        # Wrap the moov in BytesIO for independent seeking
        moov = io.BytesIO(self.cache.read(atom.size))

        # reload the atom from the fixed stream
        atom = self.read_atom(moov)

        for atom in self._find_atoms_ex(atom, moov):
            # Read either 32-bit or 64-bit offsets
            ctype, csize = ('L', 4) if atom.name == 'stco' else ('Q', 8)

            # Get number of entries
            version, entry_count = struct.unpack(">2L", moov.read(8))
            entries_pos = moov.tell()
            struct_fmt = ">%s%s" % (entry_count, ctype)

            # Read entries
            entries = struct.unpack(struct_fmt, moov.read(csize * entry_count))

            # Patch and write entries
            offset_entries = map(lambda entry: entry + offset, entries)
            moov.seek(entries_pos)
            moov.write(struct.pack(struct_fmt, *offset_entries))

        return moov



    def _find_atoms_ex(self, parent_atom, datastream):
        """
            Yield either "stco" or "co64" Atoms from datastream.
            datastream will be 8 bytes into the stco or co64 atom when the value
            is yielded.
            It is assumed that datastream will be at the end of the atom after
            the value has been yielded and processed.
            parent_atom is the parent atom, a 'moov' or other ancestor of CO
            atoms in the datastream.
        """
        stop = parent_atom.position + parent_atom.size

        while datastream.tell() < stop:
            try:
                atom = self.read_atom(datastream)
            except:
                raise FileIsCorrupt("Error reading next atom!")

            if atom.name in CHILD_ANCESTORS:
                # Known ancestor atom of stco or co64, search within it
                for res in self._find_atoms_ex(atom, datastream):
                    yield res
            elif atom.name in DESIRED_CHILDREN:
                yield atom
            else:
                # Ignore this atom, seek to the end of it.
                datastream.seek(atom.position + atom.size)

    def reset(self, stream=None):
        self.stream = stream
        self._stream_pos = 0
        self.cache = io.BytesIO()
        self.indexes = []

    def iter_stream(self):
        CHUNK_SIZE = -1
        while CHUNK_SIZE:
            chk = self.stream.read(self.BUFFER_SIZE)
            CHUNK_SIZE = len(chk)
            if CHUNK_SIZE:
                yield chk

    def iter_atom_data(self, blacklist=None, whitelist=None):
        blacklist = set(blacklist or {})
        whitelist = set(whitelist or {})
        if len(blacklist):
            filter_ = lambda x: x.name not in blacklist
        else:
            filter_ = lambda x: x.name in whitelist

        for atom in filter(filter_, self.indexes):
            self.cache.seek(atom.position)
            yield self.cache.read(atom.size)

    def from_file(self, input_file, output_file=None):
        """ @input_file: (#str) the input filename relative to the current
                working directory
            @output_file: (#str) optional file to output to. If no filename is
                provided, this method will yield :class:io.BytesIO chunks.
        """
        with open(humanfriendly.parse_path(input_file), 'rb') as f:
            for output in self.from_stream(f):
                yield output

    def from_stream(self, stream, output_file=None):
        """ @stream: An object with a 'read' attribute which returns a
                bytes-like object, 'seek' is NOT required
            @output_file: (#str) optional file to output to. If no filename is
                provided, this method will yield :class:io.BytesIO chunks.
        """
        self.reset(stream)
        CHUNK_SIZE = -1
        SKIP_TO = 0
        top_level_atoms = set()
        ccount = 0

        while CHUNK_SIZE:
            # Read the chunk from the stream
            chk = self.stream.read(self.BUFFER_SIZE)
            # Write chunk to io.BytesIO object so it is guaranteed to be
            # seekable
            CHUNK_SIZE = len(chk)
            CHUNK = io.BytesIO(chk)
            # Write the chunk to cached data, necessary for seeking outside of
            # the stream (which may or may not be seekable)
            self.cache.write(chk)
            # Skip bytes we don't need to process because we know they are part
            # of existing atoms
            chunk_skip_to = 0
            if self.cache.tell() < SKIP_TO:
                self._stream_pos = self.cache.tell()
                continue
            else:
                chunk_skip_to = SKIP_TO - self._stream_pos
            # Patch bytes in order to make sure we have at least 16 to pass to
            # the index finder
            if 0 < chunk_skip_to < 16:
                chk = self.stream.read(16 - chunk_skip_to)
                CHUNK.seek(0, os.SEEK_END)
                CHUNK.write(chk)
                CHUNK_SIZE += len(chk)
                self.cache.write(chk)
            # Seek the chunk to its proper starting point, that is, where the
            # next atom is supposed to begin
            CHUNK.seek(chunk_skip_to)
            # Find the indexes within this chunk
            for index in self.get_indexes(CHUNK, CHUNK_SIZE):
                self.indexes.append(index)
                SKIP_TO = index.position + index.size
            self._stream_pos = self.cache.tell()
            # Quit searching if we've already found the indexes that we care
            # about.
            top_level_atoms = set(item.name for item in self.indexes)
            if REQUIRED_ATOMS.issuperset(top_level_atoms) or \
               top_level_atoms.issuperset(REQUIRED_ATOMS):
                break
            elif not bool(ATOMS & top_level_atoms) and \
                 self._stream_pos >= 5 * 1024 * 1024:
                raise FileIsUnsupported('This file does not appear to be '
                                        'atom-based.')

        if not REQUIRED_ATOMS.issubset(top_level_atoms):
            raise FileIsCorrupt("A required atom was not found.")

        mdat_pos = float('inf')
        free_size = 0

        for atom in self.indexes:
            if atom.name == "moov":
                moov_atom = atom
                # If we haven't read the full atom yet, do so
                if atom.size + atom.position > self._stream_pos:
                    stream = self.stream.read(atom.size + atom.position -
                                              self._stream_pos)
                    self.cache.write(stream)
                    self._stream_pos = self.cache.tell()
            elif atom.name == "mdat":
                mdat_pos = atom.position
            elif atom.name == "free" and atom.position < mdat_pos:
                # This free atom is before the mdat!
                free_size += atom.size
            elif atom.name == "\x00\x00\x00\x00" and atom.position < mdat_pos:
                # This is some strange zero atom with incorrect size
                free_size += 8

        # Offset to shift positions
        offset =- free_size
        if moov_atom.position > mdat_pos:
            # moov is in the wrong place, shift by moov size
            offset += moov_atom.size

        if offset <= 0:
            # No free atoms to process and moov is correct, we are done so
            # just yield the file contents
            self.cache.seek(0)
            yield self.cache.read()

            # Yield the rest of the stream
            for data in self.iter_stream():
                yield data
        else:
            # Check for compressed moov atom
            self._check_compressed(moov_atom)

            # Yield ftype
            for data in self.iter_atom_data(whitelist={"ftyp"}):
                yield data

            # Read and fix moov
            moov = self._patch_moov(moov_atom, offset)
            moov = moov.getvalue()
            # Yield the moov
            yield moov

            # Yield the rest of the atoms
            for data in self.iter_atom_data(blacklist={"ftyp", "moov", "free"}):
                yield data

            # Yield the rest of the stream
            for data in self.iter_stream():
                yield data



# -*- coding: utf-8 -*-
from ...cli import *
from .._protocol import Protocol


__all__ = 'CryptoIn',


class CryptoIn(Protocol):
    """
    `AES-encrypted stream reading protocol`
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    @key: (#str) Set the AES decryption key binary block from given hexadecimal
        representation.
    @iv: (#str) Set the AES decryption initialization vector binary block from
        given hexadecimal representation.
    """

    def __init__(self, endpoint, *opt, **opts):
        super().__init__(
            *opt,
            Opt('i', 'endpoint', value='crypto+' + endpoint),
            Opt('key'),
            Opt('iv')
            **opts
        )



# -*- coding: utf-8 -*-
from ..cli import *
from ._container import *


__all__ = 'Ogg',


class Ogg(Container):
    """
    `OGG container format`
    ~~~~~~~~~~~~~~~~~~~~~~
    @page_duration: (#int) Preferred page duration, in microseconds. The muxer
        will attempt to create pages that are approximately duration
        microseconds long. This allows the user to compromise between seek
        granularity and container overhead. The default is 1 second. A value
        of 0 will fill all segments, making pages as large as possible. A value
        of 1 will effectively use 1 packet-per-page in most situations, giving
        a small seek granularity at the cost of additional container overhead.
    @serial_offset: (#int) Serial value from which to set the streams serial
        number. Setting it to different and sufficiently large values ensures
        that the produced ogg files can be safely chained.

    """
    OPT = [
        Opt('f', 'format', value='ogg'),
        IntOpt('page_duration'),
        IntOpt('serial_offset')
    ]



# -*- coding: utf-8 -*-
"""
`Opus`
============
libopus wrapper for FFMPEG
"""
from ...cli import *
from .._codec import *


__all__ = 'Opus',


class Opus(AudioCodec):
    OPT = [Opt('c:a', 'codec', value='libopus')]



# -*- coding: utf-8 -*-
"""
`AAC`
============
libfdk_aac wrapper for FFMPEG
"""
from ...cli import *
from .._codec import *


__all__ = 'Aac',


class Aac(AudioCodec):
    """
    The libfdk-aac library is based on the Fraunhofer FDK AAC code from the
    Android project. This encoder is considered to produce output on par or
    worse at 128kbps to the the native FFmpeg AAC encoder but can often produce
    better sounding audio at identical or lower bitrates and has support for
    the AAC-HE profiles.

    VBR encoding, enabled through the vbr or flags +qscale options, is
    experimental and only works with some combinations of parameters.


    ``Advanced Audio Coding (AAC) encoder``
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    @profile: (#str) Set audio profile: 'aac_low', 'aac_he', 'aac_he_v2',
        'aac_ld', 'aac_eld'. If not specified it is set to ‘aac_low’.
    @afterburner: (#bool) Enable afterburner feature if set to 1, disabled if
        set to 0. This improves the quality but also the required processing
        power.
    @eld_sbr: (#bool) Enable SBR (Spectral Band Replication) for ELD if set to
        1, disabled if set to 0. Default value is 0.
    @signaling: (#str) Set SBR/PS signaling style: 'default', 'implicit',
        'explicit_sbr', 'explicit_hierarchical'
    @latm: (#bool) Output LATM/LOAS encapsulated data if set to 1, disabled if
        set to 0.
    @header_period: (#int) Set StreamMuxConfig and PCE repetition period
        (in frames) for sending in-band configuration buffers within LATM/LOAS
        transport layer.
    @vbr: (#int 1-5) Set VBR mode, from 1 to 5. 1 is lowest quality (though
        still pretty good) and 5 is highest quality. A value of 0 will disable
        VBR, and CBR (Constant Bit Rate) is enabled. Currently only the
        ‘aac_low’ profile supports VBR encoding.
    @flags: (#str) '+qscale' Enable fixed quality, VBR (Variable Bit Rate)
        mode. Note that VBR is implicitly enabled when the vbr value is
        positive.
    """
    OPT = [
        Opt('c:a', 'codec', value='libfdk_aac'),
        OneOfOpt(
            'profile:a',
            'profile',
            choices=('aac_low', 'aac_he', 'aac_he_v2', 'aac_ld')
        ),
        FlagsOpt('flags'),
        EnableOpt('afterburner'),
        EnableOpt('eld_sbr'),
        OneOfOpt(
            'signaling',
            choices=(
                'default',
                'implicit',
                'explicit_sbr',
                'explicit_hierarchical'
            )
        ),
        EnableOpt('latm'),
        IntOpt('header_period'),
        BoundedOpt('vbr', lower=1, upper=5)
    ]



# -*- coding: utf-8 -*-
from ..cli import *


__all__ = 'Container',


DEFAULT_OPT = [
    FlagsOpt('avioflags', 'flags'),
    IntOpt('probesize', 'probe_size'),
    IntOpt('packetsize', 'packet_size'),
    FlagsOpt('fflags'),
    EnableOpt('seek2any', 'seek_to_any'),
    IntOpt('analyzeduration', 'analyze_duration'),
    Opt('cryptokey', 'crypto_key'),
    IntOpt('indexmem', 'index_memory'),
    IntOpt('rtbufsize', 'rt_bufsize'),
    FlagsOpt('fdebug'),
    IntOpt('max_delay'),
    IntOpt('fpsprobesize', 'fps_probesize'),
    IntOpt('audio_prelaod'),
    IntOpt('chunk_duration'),
    ByteOpt('chunk_size'),
    FlagsOpt('err_detect'),
    IntOpt('max_interleave_delta'),
    EnableOpt('use_wallclock_as_timestamps'),
    OneOfOpt('avoid_negative_ts',
             choices=('make_non_negative', 'make_zero', 'auto',
                      'disabled')),
    EnableOpt('skip_initial_bytes'),
    EnableOpt('correct_ts_overflow'),
    EnableOpt('flush_packets'),
    TimeOpt('output_ts_offset'),
    Opt('dump_separator')
]


class Container(OptSet):
    """
    `Container format options`
    ~~~~~~~~~~~~~~~~~~~~~~~~~~
    @avioflags: (flags)
    @probe_size (probesize): (#int) Set probing size in bytes, i.e. the size of
        the data to analyze to get stream information. A higher value will
        enable detecting more information in case it is dispersed into the
        stream, but will increase latency. Must be an integer not lesser than
        32. It is 5000000 by default.
    @packet_size (packetsize): (#int) Set packet size.
    @fflags: (flags) Set format flags.
    @seek_to_any (seek2any): (#bool) Allow seeking to non-keyframes on demuxer
        level when supported if set to 1. Default is 0.
    @analyze_duration (analyzeduration): (#int) Specify how many microseconds
        are analyzed to probe the input. A higher value will enable detecting
        more accurate information, but will increase latency. It defaults to
        5,000,000 microseconds = 5 seconds.
    @crypto_key (cryptokey): (#str) Set decryption key.
    @index_memory (indexmem): (#int) Set max memory used for timestamp index
        (per stream).
    @rt_bufsize (rtbufsize): (#int) Set max memory used for buffering real-time
        frames.
    @fdebug: (flags) Print specific debug info.
    @max_delay: (#int) Set maximum muxing or demuxing delay in microseconds.
    @fps_probesize (fpsprobesize): (#int) Set number of frames used to
        probe fps.
    @audio_preload: (#int) Set microseconds by which audio packets should be
        interleaved earlier.
    @chunk_duration: (#int) Set microseconds for each chunk.
    @chunk_size: (#int) Set size in bytes for each chunk.
    @err_detect: (flags) Set error detection flags. f_err_detect is deprecated
        and should be used only via the ffmpeg tool.
    @max_interleave_delta: (#int) Set maximum buffering duration for
        interleaving. The duration is expressed in microseconds, and defaults
        to 1000000 (1 second).
    @use_wallclock_as_timestamps: (#bool) Use wallclock as timestamps if set
        to 1. Default is 0.
    @avoid_negative_ts: (#str) 'make_non_negative', 'make_zero', 'auto',
        'disabled'
    @skip_initial_bytes: (#bool) Set number of bytes to skip before reading
        header and frames if set to 1. Default is 0.
    @correct_ts_overflow: (#bool) Correct single timestamp overflows if set
        to 1. Default is 1.
    @flush_packets: (#bool) Flush the underlying I/O stream after each packet.
        Default 1 enables it, and has the effect of reducing the latency; 0
        disables it and may slightly increase performance in some cases.
    @output_ts_offset: (#str (time)) Set the output time offset. Offset must be
        a time duration specification, see the Time duration section in the
        ffmpeg-utils(1) manual.
    @dump_separator: (#str) Separator used to separate the fields printed on
        the command line about the Stream parameters.
    """
    OPT = []

    def __init__(self, *opt, **opts):
        super().__init__(*(DEFAULT_OPT + self.OPT + list(opt)), **opts)

    def __call__(self, *args, **kwargs):
        return self.__class__(*args, **kwargs)



# -*- coding: utf-8 -*-
from ..cli import *
from ._container import *


__all__ = 'Chromaprint',


class Chromaprint(Container):
    """
    This muxer feeds audio data to the Chromaprint library, which generates a
    fingerprint for the provided audio data. It takes a single signed
    native-endian 16-bit raw audio stream.


    ``Chromaprint container format``
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    @silence_threshold: (#int -1-32767) Threshold for detecting silence, ranges
        from 0 to 32767. -1 for default (required for use with the AcoustID
        service).
    @algorithm: (#int) Algorithm index to fingerprint with.
    @fp_format: (#str) Format to output the fingerprint as: 'raw', 'compressed',
        'base64'
    """
    OPT = [
        Opt('f', 'format', value='chromaprint'),
        BoundedOpt('silence_threshold', lower=-1, upper=32767),
        OneOfOpt('fp_format', choices=('raw', 'compressed', 'base64'))
    ]



# -*- coding: utf-8 -*-
from .cli import *
from .format import Container
from .probe import Probe


__all__ = 'Output', 'Video', 'Audio', 'Subtitle'


DEFAULT_OUTPUT_OPT = [
    Opt('map_metadata'),
    FloatOpt('t', 'duration'),
    TimeOpt('to', 'stop_time'),
    TimeOpt('ss', 'start_time'),
    TimeOpt('sseof', 'rstart_time'),
    Opt('target'),
    KeyValOpt('program'),
    KeyValOpt('metadata'),
    IntOpt('frames'),
    Opt('filter_script'),
    Opt('reinit_filter'),
    ByteOpt('fs', 'max_size'),
    Flag('seek_timestamp'),
    Flag('apad', 'audio_pad'),
    Flag('discard'),
    Flag('disposition')
]


class Output(OptSet):
    '''
    -map_metadata       set metadata information of outfile from infile
    -t duration         record or transcode "duration" seconds of audio/video
    -to time_stop       record or transcode stop time
    -fs limit_size      set the limit file size in bytes
    -ss time_off        set the start time offset
    -sseof time_off     set the start time offset relative to EOF
    -seek_timestamp     enable/disable seeking by timestamp with -ss
    -metadata string=string  add metadata (key=value)
    -program title=string:st=number...  add program with specified streams
    -target type        specify target file type
    -apad               audio pad
    -frames number      set the number of frames to output
    -filter_script filename  read stream filtergraph description from a file
    -reinit_filter      reinit filtergraph on input parameter changes
    -discard            discard
    -disposition        disposition
    '''
    OPT = []

    def __init__(self, *opt, **opts):
        """ @codec: (:class:transcend.coding.codecs.Codec) codec settings
            @format: (:class:transcend.coding.format.Container)
                format settings
            @output: (#str or :class:transcend.coding.protocols.Protocol)
                output protocol
        """
        self.codec = None
        self.format = None
        self.output = None

        if len(opts):
            for prop in ('codec', 'format', 'output'):
                try:
                    val = opts[prop]
                    del opts[prop]
                    setattr(self, prop, val)
                except KeyError:
                    pass

        super().__init__(*(list(opt) + DEFAULT_OUTPUT_OPT + self.OPT))
        self.set(**opts)

    @property
    def live_opts(self):
        if self.codec is not None:
            for cmd in self.codec.live_opts:
                yield cmd
        if self.format is not None:
            for cmd in self.format.live_opts:
                yield cmd
        for cmd in super(Output, self).live_opts:
            yield cmd
        if self.output is not None:
            if isinstance(self.output, str):
                yield self.output
            else:
                for cmd in self.output.live_opts:
                    yield cmd

    def save_to(self, output):
        self.output = output


class Video(Output):
    '''
    -vframes number     set the number of video frames to output
    -vn                 disable video
    -an                 disable audio
    -vcodec codec       force video codec ('copy' to copy stream)
    -timecode hh:mm:ss[:;.]ff  set initial TimeCode value.
    -pass n             select the pass number (1 to 3)
    -vf filter_graph    set video filters
    -dn                 disable data
    -pix_fmt            set pixel format
    '''
    OPT = [
        BoundedOpt('pass', 'pass_n', lower=1, upper=2),
        IntOpt('vframes', 'frames'),
        Opt('vcodec'),
        Flag('vn', 'disable'),
        Flag('an', 'disable_audio'),
        TimeOpt('timecode', 'timecode'),
        KeyValOpt('filter:v', 'filters'),
        KeyValOpt('filter_complex', 'complex_filters'),
        Flag('dn', 'disable_data'),
        Flag('sn', 'disable_subtitles'),
        Opt('pix_fmt', 'pixel_format')
    ]

    def __init__(self, *opt, **opts):
        """ @audio: (:class:Audio) audio track settings
            @codec: (:class:transcend.coding.codecs.Codec) codec settings
            @format: (:class:transcend.coding.format.Container)
                format settings
            @output: (#str or :class:transcend.coding.protocols.Protocol)
                output protocol
        """
        self.audio = None
        self.subtitles = None

        if len(opts):
            for prop in ('audio', 'subtitles'):
                try:
                    val = opts[prop]
                    del opts[prop]
                except KeyError:
                    val = None
                setattr(self, prop, val)
        super().__init__(*opt, **opts)

    def set_audio(self, audio):
        self.audio = audio

    @property
    def live_opts(self):
        if self.audio is not None:
            for cmd in self.audio.live_opts:
                yield cmd
        if self.subtitles is not None:
            for cmd in self.subtitles.live_opts:
                yield cmd
        for cmd in super(Video, self).live_opts:
            yield cmd

    def resize(self):
        '''https://trac.ffmpeg.org/wiki/Scaling%20(resizing)%20with%20ffmpeg'''


class Audio(Output):
    '''
    -aframes number     set the number of audio frames to output
    -an                 disable audio
    -vn                 disable video
    -acodec codec       force audio codec ('copy' to copy stream)
    -af filter_graph    set audio filters
    '''
    OPT = [
        IntOpt('aframes', 'frames'),
        Opt('acodec'),
        Flag('an', 'disable'),
        Flag('vn', 'disable_video'),
        Flag('sn', 'disable_subtitles'),
        KeyValOpt('filter:a', 'filters')
    ]


class Subtitle(Output):
    '''
    -s size             set frame size (WxH or abbreviation)
    -sn                 disable subtitle
    -scodec codec       force subtitle codec ('copy' to copy stream)
    -stag fourcc/tag    force subtitle tag/fourcc
    -fix_sub_duration   fix subtitles duration
    -canvas_size size   set canvas size (WxH or abbreviation)
    -spre preset        set the subtitle options to the indicated preset
    '''
    OPT = [
        SizeOpt('s', 'size'),
        Flag('sn', 'disable'),
        Flag('vn', 'disable_video'),
        Flag('an', 'disable_audio'),
        Opt('c:s', 'codec'),
        Opt('stag', 'tag'),
        Flag('fix_duration'),
        SizeOpt('canvas_size'),
        Opt('spre', 'preset')
    ]



# -*- coding: utf-8 -*-
"""
`Vorbis`
============
libvorbis wrapper for FFMPEG
"""
from ...cli import *
from .._codec import *


__all__ = 'Vorbis',


class Vorbis(AudioCodec):
    """
    @iblock: (#float) Set noise floor bias for impulse blocks. The value is a
        float number from -15.0 to 0.0. A negative bias instructs the encoder
        to pay special attention to the crispness of transients in the encoded
        audio. The tradeoff for better transient response is a higher bitrate.
    """
    OPT = [
        Opt('c:a', 'codec', value='libvorbis'),
        BoundedOpt('compression_level', cast=float, lower=-15.0, upper=0.0)
    ]



# -*- coding: utf-8 -*-
from ...cli import *
from .._container import *


__all__ = ('WebMDash', 'WebMChunk')


class WebMDash(Container):
    """
    This muxer implements the WebM DASH Manifest specification to generate the
    DASH manifest XML. It also supports manifest generation for DASH live
    streams.


    ``WebM DASH Manifest muxer options``
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    @adaptation_sets: (#str) "id=x,streams=a,b,c id=y,streams=d,e" where x and
        y are the unique identifiers of the adaptation sets and a,b,c,d and e
        are the indices of the corresponding audio and video streams. Any number
        of adaptation sets can be added using this option.
    @live: (#bool) Set this to 1 to create a live stream DASH Manifest.
        Default: 0.
    @chunk_start_index: (#int) Start index of the first chunk. This will go in
        the ‘startNumber’ attribute of the ‘SegmentTemplate’ element in the
        manifest. Default: 0.
    @chunk_duration (chunk_duration_ms): (#int) Duration of each chunk in
        milliseconds. This will go in the ‘duration’ attribute of the
        ‘SegmentTemplate’ element in the manifest. Default: 1000.
    @utc_timing_url: (#str) URL of the page that will return the UTC timestamp
        in ISO format. This will go in the ‘value’ attribute of the ‘UTCTiming’
        element in the manifest. Default: None.
    @time_shift_buffer_depth: (#int) Smallest time (in seconds) shifting
        buffer for which any Representation is guaranteed to be available.
        This will go in the ‘timeShiftBufferDepth’ attribute of the ‘MPD’
        element. Default: 60.
    @minimum_update_period: (#int) Minimum update period (in seconds) of the
        manifest. This will go in the ‘minimumUpdatePeriod’ attribute of the
        ‘MPD’ element. Default: 0.

    """
    OPT = [
        Opt('f', 'format', value='webm_dash_manifest'),
        Opt('adaptation_sets'),
        EnableOpt('live'),
        IntOpt('chunk_start_index'),
        IntOpt('chunk_duration'),
        Opt('utc_timing_url'),
        IntOpt('time_shift_buffer_depth'),
        IntOpt('minimum_update_period'),
    ]


class WebMChunk(Container):
    """
    This muxer writes out WebM headers and chunks as separate files which can
    be consumed by clients that support WebM Live streams via DASH.


    ``WebM Live Chunk muxer options``
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    @chunk_start_index: (#int) Index of the first chunk (defaults to 0).
    @audio_chunk_duration: (#int) Duration of each audio chunk in milliseconds
        (defaults to 5000).
    @header: (#str) Filename of the header where the initialization data will
        be written.
    """
    OPT = [
        Opt('f', 'format', value='webm_chunk'),
        IntOpt('chunk_start_index'),
        IntOpt('audio_chunk_duration'),
        Opt('header')
    ]



# -*- coding: utf-8 -*-
from ...cli import *
from .._protocol import Protocol


__all__ = 'AsyncIn',


class AsyncIn(Protocol):
    """
    `Asynchronous data filling wrapper protocol`
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    Fill data in a background thread, to decouple I/O operation from demux
    thread.

    async:URL
    async:http://host/resource
    async:cache:http://host/resource
    """
    def __init__(self, endpoint, *opt, **opts):
        super().__init__(
            *opt,
            Opt('i', 'endpoint', value='async:' + endpoint),
            **opts
        )



# -*- coding: utf-8 -*-
from ..cli import *
from ._protocol import Protocol


__all__ = 'UdpIn', 'UdpOut'


class _Udp(Protocol):
    """
    `User Datagram Protocol`
    """


class UdpIn(_Udp):
    def __init__(self, endpoint, *opt, **opts):
        super().__init__(
            *opt,
            Opt('i', 'endpoint', value=endpoint),
            **opts
        )


class UdpOut(_Udp):
    def __init__(self, endpoint, *opt, **opts):
        super().__init__(
            *opt,
            Opt('', 'endpoint', value=endpoint),
            **opts
        )



# -*- coding: utf-8 -*-
from ..cli import *
from ._container import *


__all__ = 'Null',


class Null(Container):
    """
    `Null container format`
    ~~~~~~~~~~~~~~~~~~~~~~~
    For debugging and benchmarking purposes.
    """
    OPT = [Opt('f', 'format', value='null')]



# -*- coding: utf-8 -*-
from ..cli import *
from ._protocol import Protocol


__all__ = 'Pipe', 'PipeIn', 'PipeOut', 'PipeErr',


class Pipe(Protocol):
    """
    `Pipe access protocol`
    ~~~~~~~~~~~~~~~~~~~~~~
    @number: (#bool) 0=stdin, 1=stdout, 2=stderr
    @blocksize: (#str) Set I/O operation maximum block size, in bytes. Default
        value is INT_MAX, which results in not limiting the requested block
        size. Setting this value reasonably low improves user termination
        request reaction time, which is valuable for files on slow medium.

    """
    def __init__(self, *opt, **opts):
        pname = 'i' if str(self.NUMBER) == '0' else ''
        
        super().__init__(
            ByteOpt('blocksize'),
            Opt(pname, 'in_out_err', value='pipe:%s' % self.NUMBER),
            *opt,
            **opts
        )


class PipeIn(Pipe):
    NUMBER = 0


class PipeOut(Pipe):
    NUMBER = 1


class PipeErr(Pipe):
    NUMBER = 2



from multiprocessing import cpu_count
from ...codec import Theora
from ...output import Video, Audio
from ...format import WebM
from ..audio import vorbis



class BaseTheora(Video):

    def __init__(self, input=None, **opts):
        super(BaseTheora, self).__init__(
            pixel_format='yuv420p',
            codec=Theora(threads=cpu_count()),
            format=WebM(),
            **opts)


class AnyDevice(BaseTheora):

    def __init__(self, input=None, **opts):
        super(AnyDevice, self).__init__(
            audio=vorbis.LoFiAudio,
            **opts)
        self.codec.set(size='424x240',
                       bitrate='576k',
                       max_bitrate='576k',
                       buffer_size='864k')


class AnyDevice360p(AnyDevice):

    def __init__(self, input=None, **opts):
        super(AnyDevice360p, self).__init__(
            **opts)
        self.codec.set(size='640x360',
                       bitrate='896k',
                       max_bitrate='896k',
                       buffer_size='1344k')


class Theora360p(BaseTheora):

    def __init__(self, input=None, **opts):
        super(Theora360p, self).__init__(
            audio=vorbis.LoFiAudio,
            **opts)
        self.codec.set(size='640x360',
                       bitrate='896k',
                       max_bitrate='896k',
                       buffer_size='1344k',
                       crf=16)


class Theora480p(BaseTheora):

    def __init__(self, input=None, **opts):
        super(Theora480p, self).__init__(
            audio=vorbis.MidFiAudio,
            **opts)
        self.codec.set(size='848x480',
                       bitrate='1536k',
                       max_bitrate='1536k',
                       buffer_size='2304k')


class Theora576p(BaseTheora):
    def __init__(self, input=None, **opts):
        super(Theora576p, self).__init__(
            audio=vorbis.MidFiAudio,
            **opts)
        self.codec.set(size='960x576',
                       bitrate='2176k',
                       max_bitrate='2176k',
                       buffer_size='3264k')


class Theora720p(BaseTheora):
    def __init__(self, input=None, **opts):
        super(Theora720p, self).__init__(
            audio=vorbis.HiFiAudio,
            **opts)
        self.codec.set(size='1280x270',
                       bitrate='2176k',
                       max_bitrate='2176k',
                       buffer_size='3264k')


class Theora1080p(BaseTheora):
    def __init__(self, input=None, **opts):
        super(Theora1080p, self).__init__(
            audio=vorbis.HiFiAudio,
            **opts)
        self.codec.set(size='1920x1080',
                       bitrate='7552k',
                       max_bitrate='7552k',
                       buffer_size='11328k')


class Theora1440p(BaseTheora):
    def __init__(self, input=None, **opts):
        super(Theora1440p, self).__init__(
            audio=vorbis.HiFiAudio,
            **opts)
        self.codec.set(size='2560x1440',
                       bitrate='18M',
                       max_bitrate='18M',
                       buffer_size='30.48M')

Theora2k = Theora1440p


class Theora2160p(BaseTheora):
    def __init__(self, input=None, **opts):
        super(Theora2160p, self).__init__(
            audio=vorbis.HiFiAudio,
            **opts)
        self.codec.set(size='3840×2160',
                       bitrate='55M',
                       max_bitrate='55M',
                       buffer_size='82.5M')

Theora4k = Theora2160p



#!/usr/bin/python3 -S
import os
import sys
import uuid
from setuptools import setup


def compat(min_version):
    ''' @min_version: (#str) minimum version number formatted like |2.7|
            or |2.7.6|
        -> (#bool) |True| if the system version is at least @min_version
    '''
    return sys.version_info >= tuple(map(int, min_version.split('.')))


PKG = 'transcend'
PKG_NAME = 'transcend-py'
PKG_VERSION = '0.1.6'

pathname = os.path.dirname(os.path.realpath(__file__))


def parse_requirements(filename):
    """ load requirements from a pip requirements file """
    lineiter = (line.strip() for line in open(filename))
    return (line for line in lineiter if line and not line.startswith("#"))


install_reqs = parse_requirements(pathname + "/requirements.txt")

setup(
    name=PKG_NAME,
    version=PKG_VERSION,
    description='High performance video and audio transcoding for aliens.',
    author='Jared Lunde',
    author_email='jared.lunde@gmail.com',
    url='https://github.com/jaredlunde/transcend',
    license="MIT",
    install_requires=list(install_reqs),
    packages=[
        'transcend',
        'transcend.codec',
        'transcend.codec.audio',
        'transcend.codec.subtitle',
        'transcend.codec.video',
        'transcend.format',
        'transcend.format.streaming',
        'transcend.preset',
        'transcend.preset.audio',
        'transcend.preset.subtitle',
        'transcend.preset.video',
        'transcend.preset.video.streaming',
        'transcend.protocol',
        'transcend.protocol.input',
        'transcend.protocol.output',
        'transcend.utils',
    ]

)



# -*- coding: utf-8 -*-
from ..cli import *
from ._protocol import Protocol


__all__ = 'FtpIn', 'FtpOut'


FTP_OPT = [
    IntOpt('timeout'),
    Opt('ftp-anonymous-password', 'anonymous_password'),
    EnableOpt('ftp-write-seekable', 'write_seekable'),
    EnableOpt('follow')
]


class _Ftp(Protocol):
    """
    `File transfer protocol`
    ~~~~~~~~~~~~~~~~~~~~~~~~
    @timeout: (#int) Set timeout in microseconds of socket I/O operations used
        by the underlying low level operation. By default it is set to -1,
        which means that the timeout is not specified.
    @anonymous_password (ftp-anonymous-password): (#str) Password used when
        login as anonymous user. Typically an e-mail address should be used.
    @write_seekable (ftp-write-seekable): (#bool) Control seekability of
        connection during encoding. If set to 1 the resource is supposed to be
        seekable, if set to 0 it is assumed not to be seekable. Default value
        is 0.
    @follow: (#bool) If set to 1, the protocol will retry reading at the end
        of the file, allowing reading files that still are being written. In
        order for this to terminate, you either need to use the rw_timeout
        option, or use the interrupt callback (for API users).
    """
    OPT = FTP_OPT


class FtpIn(_Ftp):
    def __init__(self, endpoint, *opt, **opts):
        super().__init__(
            *opt,
            Opt('i', 'endpoint', value=endpoint),
            **opts
        )


class FtpOut(_Ftp):
    def __init__(self, endpoint, *opt, **opts):
        super().__init__(
            *opt,
            Opt('', 'endpoint', value=endpoint),
            **opts
        )



# -*- coding: utf-8 -*-
from ...cli import *
from .._container import *


__all__ = 'Dash',


# https://ffmpeg.org/ffmpeg-formats.html#dash-1

class Dash(Container):
    """
    Dynamic Adaptive Streaming over HTTP (DASH) muxer that creates segments
    and manifest files according to the MPEG-DASH standard
    ISO/IEC 23009-1:2014.

    It creates a MPD manifest file and segment files for each stream.

    The segment filename might contain pre-defined identifiers used with
    SegmentTemplate as defined in section 5.3.9.4.4 of the standard.
    Available identifiers are "$RepresentationID$", "$Number$",
    "$Bandwidth$" and "$Time$".

    Example
    ..
        ffmpeg -re -i <input> -map 0 -map 0 -c:a libfdk_aac -c:v libx264
        -b:v:0 800k -b:v:1 300k -s:v:1 320x170 -profile:v:1 baseline
        -profile:v:0 main -bf 1 -keyint_min 120 -g 120 -sc_threshold 0
        -b_strategy 0 -ar:a:1 22050 -use_timeline 1 -use_template 1
        -window_size 5 -adaptation_sets "id=0,streams=v id=1,streams=a"
        -f dash /path/to/out.mpd
    ..
    """
    OPT = [
        Opt('f', 'format', value='dash'),
        IntOpt('min_seg_duration'),
        IntOpt('window_size'),
        IntOpt('extra_window_size'),
        EnableOpt('remove_at_exit'),
        EnableOpt('use_template'),
        EnableOpt('use_timeline'),
        EnableOpt('single_file'),
        Opt('single_file_name'),
        Opt('init_seg_name'),
        Opt('media_seg_name'),
        Opt('utc_timing_url'),
        KeyValOpt('adaptation_sets'),
    ]



# -*- coding: utf-8 -*-
from ..cli import *
from ._container import *


__all__ = 'Asf',


class Asf(Container):
    """
    `Advanced Systems Format container format`
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    @packet_size: (#str) Set the muxer packet size. By tuning this setting you
        may reduce data fragmentation or muxer overhead depending on your
        source. Default value is 3200, minimum is 100, maximum is 64k.
    """
    OPT = [
        Opt('f', 'format', value='asf'),
        BitOpt('packet_size')
    ]
